/*
    Main.c
    This is the main source code file for the game of Pong.
                
    Version 1.0   May 2007

    Copyright 2007 Geoff Graham - http://geoffg.net
    This program is free software: you can redistribute it and/or modify it under the terms of the GNU General
    Public License as published by the Free Software Foundation, either version 2 of the License, or (at your
    option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.  You should have received a copy of the GNU General Public License along with this program.
    If not, see <http://www.gnu.org/licenses/>.
*/

//////////////////////////////////////////////////////////////////////////////////////////////////////////
// Program summary:
//
// - Timer 0 divides the processor clock (16MHz) to generate an interrupt at exactly the PAL horizontal
//   sync rate (64 uS).  This interrupt is handled (in assembly language) to generate the horiz/vert sync
//   pulses, video and sound.
//
// - Video is generated by reading through a buffer (LBUF) during the interrupt and putting the contents
//   of the buffer out on the video output pin of the processor.
//
// - Each nibble (4 bits) of a byte in the buffer represents one pixel.  The nibbles are swapped and the
//   whole byte is written to the output port, then the nibbles swapped again and written again.  This is
//   two pixels.  The same is done for the next byte in the buffer (2 pixels) and so on for 40 bytes.
//   Therefor the horizontal line is 80 pixels with the top nibble representing the first pixel.
//
// - The main calculation routine (written in C) runs continuously calculating ball and bat positions
//   and updating variables shared by both the main program and the interrupt routine.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////

#Define 16F688

//////////////////////////////////////////////////////////////////////////////////////////////////////////
// The processor specific defines for the 16F88 chip
// NOTE THAT THE SETTINGS FOR THE 16F88 HAVE NOT BEEN FULLY TESTED - EXPECT ERRORS
#ifdef 16F88
    #include <16F88.h>
    #fuses XT,NOWDT,NOPROTECT,MCLR,NOBROWNOUT, LVP, NODEBUG
    // Clock Related Fuses (select one):
    //        LP  Low-Power Crystal
    //        XT  Crystal/Resonator
    //        HS  High-Speed Crystal/Resonator
    //        RC  External Resistor-Capacitor (RC) with FOSC/4 output on RA6
    //        RC_IO  External Resistor-Capacitor with I/O on RA6
    //        INTRC - Internal Oscillator with FOSC/4 output on RA6 and I/O on RA7
    //        INTRC_IO - Internal Oscillator with I/O on RA6 and RA7
    //        EC_IO  External Clock with I/O on RA6.
    // Int/Ext Clock Switchover:     IESO,NOIESO
    // Fail-Safe Clock Monitor:      FCMEN,NOFCMEN
    // Watch Dog Timer:              WDT, NOWDT
    // Protect Data Memory:          CPD,NOCPD
    // Protect Code:                 PROTECT,NOPROTECT
    // Use MCLR Pin for Reset:       MCLR, NOMCLR
    // Power-up Timer:               PUT,NOPUT
    // Brown-out Detect:             BROWNOUT,NOBROWNOUT
    // Low-Volt Programming Enable:  LVP, NOLVP
    // In-Circuit Debugger Mode:     DEBUG,NODEBUG
    // CCP1 Pin Selection:           CCPB0, CCPB3

    #device ADC=8
    #use delay(xtal=16M)
    
    // processor specific hardware registers
    #byte indf = 0                  // Indirect register
    #byte tmr0 = 1                  // timer count register
    #byte status = 3                // The status register
    #byte fsr = 4                   // Index to RAM
    #byte port_a = 5                // Port A register
    #byte port_b = 6                // Port B register
    #byte pclath = 0x0a             // Top bits of the program counter
    #byte intcon = 0x0b             // Basic interrupt control
    #byte plr1 = 0x0c               // Interrupt flags
    #byte option_reg = 0x81         // I/O ports options
    #byte ple1 = 0x8c               // Interrupt enables
    #byte spbrg = 0x99              // Baud rate

    // where to locate assembly language memory. 
    #define VAR_START 0x20          // Must be in Bank 00 and be free to approx VAR_START+60

    // input/out pins defined
    #define TRIS_A  0b11111000      // I/O pins for port A
    #define TRIS_B  0b11111011      // I/O pins for port B
    #define ANALOG_PINS sAN3||sAN4  // ADC input ports
    
    // assembly language constants for ports managed in that language
    #define SYNC_PORT   port_a      // Sync pulses out on this port
    #define SYNC_PIN    0           // Sync pulses out on this pin of Port A
    
    #define VIDEO_PORT  port_b      // Video pulses out on this port
    #define VIDEO_PIN   2           // Video pulses out on this pin of Port B

    #define AUDIO_PORT  port_a      // Audio out on this port
    #define AUDIO_BIT   0b00000100  // Audio out on this bit of Port A
    
    // constants for ports managed in the C language
    #define LBAT_CHAN   3           // Analog channel for the left bat
    #define RBAT_CHAN   4           // Analog channel for the right bat
    #define RUNSW       PIN_B0      // Input for the run switch
    #define BAT5V       PIN_B1      // Source of 5 volts for the bat potentiometers
    #define SKILLSW     PIN_B7      // Input for the game difficulty switch

    //
    //                           -----
    //            SOUND OUT  1  |     | 18
    //             LEFT BAT  2  |  1  | 17  SYNC OUT
    //            RIGHT BAT  3  |  6  | 16  XTAL 1
    //  ISCP Pin 1 &  RESET  4  |  F  | 15  XTAL 2
    //               GROUND  5  |  8  | 14  +5 Volts
    //               RUN SW  6  |  8  | 13  SKILL SW & ISCP Pin 4
    //                       7  |     | 12  ISCP Pin 5
    //            VIDEO OUT  8  |     | 11
    //           ISCP Pin 6  9  |     | 10
    //                           -----
    //

    // processor specific constants
    #define C_FLAG      0           // Carry flag in the status register
    #define Z_FLAG      2           // Zero flag in the status register
    #define RBPU        7           // Weak pullups flag in option_reg

#endif


//////////////////////////////////////////////////////////////////////////////////////////////////////////
// The processor specific defines for the 16F688 chip
#ifdef 16F688
    #include <16F688.h>
    #fuses XT,NOWDT,NOCPD,NOPROTECT,NOMCLR,NOBROWNOUT
    // Clock Related Fuses (select one):
    //        LP  Low-Power Crystal
    //        XT  Crystal/Resonator
    //        HS  High-Speed Crystal/Resonator
    //        RC  External Resistor-Capacitor (RC) with FOSC/4 output on RA4
    //        RC_IO  External Resistor-Capacitor with I/O on RA6
    //        INTRC - Internal Oscillator with FOSC/4 output on RA4 and I/O on RA5
    //        INTRC_IO - Internal Oscillator with I/O on RA4 and RA5
    //        EC_IO  External Clock with I/O on RA4.
    // Int/Ext Clock Switchover:     IESO,NOIESO
    // Fail-Safe Clock Monitor:      FCMEN,NOFCMEN
    // Watch Dog Timer:              WDT, NOWDT
    // Protect Data Memory:          CPD,NOCPD
    // Protect Code:                 PROTECT,NOPROTECT
    // Use MCLR Pin for Reset:       MCLR, NOMCLR
    // Power-up Timer:               PUT,NOPUT
    // Brown-out Detect:             BROWNOUT,NOBROWNOUT, BROWNOUT_SW,BROWNOUT_NOSL

    #device ADC=8
    #use delay(xtal=16M)
    
    // processor specific hardware registers
    #byte indf = 0                  // Indirect register
    #byte tmr0 = 1                  // timer count register
    #byte status = 3                // The status register
    #byte fsr = 4                   // Index to RAM
    #byte port_a = 5                // Port A register
    #byte port_c = 7                // Port C register
    #byte pclath = 0x0a             // Top bits of the program counter
    #byte intcon = 0x0b             // Basic interrupt control
    #byte plr1 = 0x0c               // Interrupt flags
    #byte option_reg = 0x81         // I/O ports options
    #byte ple1 = 0x8c               // Interrupt enables
    #byte wpua = 0x95               // weak pullups config register

    // where to locate assembly language memory. 
    #define VAR_START 0x30          // Must be in Bank 00 and be free to approx VAR_START+70

    // input/out pins defined
    #define TRIS_A  0b11111011      // I/O pins for port A
    #define TRIS_C  0b11010011      // I/O pins for port C
    #define ANALOG_PINS sAN4||sAN5  // ADC input ports
    
    // assembly language constants for ports managed in that language
    #define SYNC_PORT   port_c      // Sync pulses out on this port
    #define SYNC_PIN    2           // Sync pulses out on this pin of Port A
    
    #define VIDEO_PORT  port_a      // Video pulses out on this port
    #define VIDEO_PIN   2           // Video pulses out on this pin of Port B

    #define AUDIO_PORT  port_c      // Audio out on this port
    #define AUDIO_BIT   0b00001000  // Audio out on this bit of Port A
    
    // constants for ports managed in the C language
    #define LBAT_CHAN   4           // Analog channel for the left bat
    #define RBAT_CHAN   5           // Analog channel for the right bat
    #define RUNSW       PIN_A0      // Input for the run switch
    #define BAT5V       PIN_C5      // Source of 5 volts for the bat potentiometers
    #define SKILLSW     PIN_A1      // Input for the game difficulty switch

    //
    //                           -----
    //             +5 Volts  1  |  1  | 18  GROUND
    //               XTAL 1  2  |  6  | 17  RUN SW & ISCP Pin 4
    //               XTAL 2  3  |  F  | 16  SKILL SW & ISCP Pin 5
    //           ISCP Pin 1  4  |  6  | 15  VIDEO OUT
    //      +5V OUT FOR BAT  5  |  8  | 14  LEFT BAT
    //                       6  |  8  | 13  RIGHT BAT
    //            SOUND OUT  7  |     | 12  SYNC OUT
    //                           -----
    //

    // processor specific constants
    #define C_FLAG      0           // Carry flag in the status register
    #define Z_FLAG      2           // Zero flag in the status register
    #define RBPU        7           // Weak pullups flag in option_reg

#endif

// compiler specific defines
// my code will handle the port directions
#use fast_io(A) 
#IFDEF 16F88                    
#use fast_io(B)
#ELIFDEF 16F688                 
#use fast_io(C)
#ENDIF

////////////////////////////////////////////////////////////////////////////////////////////////////////////
// program configuration defines

#define BALL_SCALE      20          // controls the speed of the game (lower number = faster)
#define BALL_START_SPEED 4          // initial speed of the ball (lower number = slower)

// the layout of the playing field
#define LBAT_X          7           // The horizontal BYTE position of the left bat from the left boundary
#define RBAT_OFFSET     9           // The horizontal BYTE distance of the right bat from the right boundary

// number of lines that each state lasts for
#define STATE0_CNT      7           // vert sync
#define STATE1_CNT      45          // blank top of the screen
#define STATE2_CNT      4           // single line for the top of the playing field
#define STATE3_CNT      220         // main playing field
#define STATE4_CNT      4           // single line for the bottom of the playing field
#define STATE5_CNT      35          // blank bottom of the screen

// playing sounds and durations
#define BEEP_WALL_FREQ  8
#define BEEP_WALL_TIME  25
#define BEEP_BAT_FREQ   16
#define BEEP_BAT_TIME   50
#define BEEP_GOAL_FREQ  32
#define BEEP_GOAL_TIME  250

// goal and bat heights
#define BAT_LARGE   0b11000000      // mask for the height of a large bat
#define BAT_STD     0b11100000      // mask for the height of a standard bat
#define BAT_SMALL   0b11110000      // mask for the height of a small bat
#define GOAL_LARGE  0b10000000      // mask for the height of a small goal
#define GOAL_STD    0b11000000      // mask for the height of a standard goal

// default skill level on first use of this program
#rom 0x2200={1}


//////////////////////////////////////////////////////////////////////////////////////////////////////////
// Define memory and variable locations that are used by the assembly language

#define LBUF_SIZE 40
#define LBUF VAR_START              // buffer for each horiz line
#reserve LBUF:(LBUF + LBUF_SIZE - 1)

int save_status;                    // where to save the status register in an interrupt
#locate save_status=LBUF + LBUF_SIZE + 1

int save_fsr;                       // where to save the fsr register in an interrupt
#locate save_fsr=LBUF + LBUF_SIZE + 2

int state;                          // the vertical field is divided into a nbr of states
#locate state=LBUF + LBUF_SIZE + 3

int line;                           // current line (within a state) being scanned
#locate line=LBUF + LBUF_SIZE + 4

int ball_X;                         // current position of the ball in the playing field
#locate ball_X=LBUF + LBUF_SIZE + 5

int ball_Y;                         // current position of the ball in the playing field
#locate ball_Y=LBUF + LBUF_SIZE + 6

int Lbat_Y;                         // current position of the left bat in the playing field
#locate Lbat_Y=LBUF + LBUF_SIZE + 7

int Rbat_Y;                         // current position of the right bat in the playing field
#locate Rbat_Y=LBUF + LBUF_SIZE + 8
#define RBAT_X LBUF_SIZE - RBAT_OFFSET

int Lbat_mask, Rbat_mask;           // used by assembly language to get the height of the bats
#locate Lbat_mask=LBUF + LBUF_SIZE + 9
#locate Rbat_mask=LBUF + LBUF_SIZE + 10

int beep_time;                      // length of beep (beep_time = mS / (beep_freq * 0.064))
#locate beep_time=LBUF + LBUF_SIZE + 11

int beep_freq;                      // frequency of beep (Hz = 7812/beep_freq)
#locate beep_freq=LBUF + LBUF_SIZE + 12

int beep_tmp;                       // temp storage used by beep routine
#locate beep_tmp=LBUF + LBUF_SIZE + 13

// clock counters used by key bounce routines and others (each count = 20mS)
int skillsw_timer, runsw_timer, auto_off_timer_1, auto_off_timer_2;                         
#locate skillsw_timer=LBUF + LBUF_SIZE + 14
#locate runsw_timer=LBUF + LBUF_SIZE + 15
#locate auto_off_timer_1=LBUF + LBUF_SIZE + 16
#locate auto_off_timer_2=LBUF + LBUF_SIZE + 17

int goal_mask, goal_top;        // used by assembly language to get the goal positions
#locate goal_mask=LBUF + LBUF_SIZE + 18
#locate goal_top=LBUF + LBUF_SIZE + 19

int practice;                       // true (all 1's) when we are in practice mode
#locate practice=LBUF + LBUF_SIZE + 20

// NOTE:  Maximum number of memory bytes that can be used is 24 (assuming LBUF_SIZE = 40)
//        Any more and you will encroach into shared space and overwrite save_w below.

int save_w;                         // where to save the w register in an interrupt
#locate save_w=0x70                 // this must be in common memory
#reserve 0x70, 0xf0, 0x170, 0x1f0   // reserve this spot in the other pages

#define LIMIT_X ((LBUF_SIZE * 2) - 1) * BALL_SCALE  // limit of the ball's travel in the horizontal
#define LIMIT_Y STATE3_CNT * BALL_SCALE             // limit of the ball's travel in the vert


//////////////////////////////////////////////////////////////////////////////////////////////////////////
// Define any functions (the stupid compiler should be multipass)
signed int16 GetBatPosition(int channel);
void SetupSkill(int skill_level);
void SetupBall();

/////////////////////////////////////////////////////////////////////////////////////////////////////////
// Define global variables used by C
signed int16 bat_size_Y;
signed int16 goal_size_Y;
signed int16 goal_top_Y;
signed int16 tball_X, tball_Y;                  // location of the ball
signed int16 speed_X, speed_Y;                  // speed of the ball
signed int16 Ltbat_Y, Rtbat_Y;                  // vertical position of the two bats
    
//////////////////////////////////////////////////////////////////////////////////////////////////////////
// The global interrupt routine
// Written in assembly language to get the timing correct during the horizontal scan
// This is used for the timer 0 interrupt which triggers at the horizontal sync rate
// There are no other interrupts so we do not have to check interrupt sources
#INT_GLOBAL
void MyGlobal() {
    
  #asm ASIS
;
; first save the sensitive processor registers
        movwf   save_w              ; store current state of w
        swapf   status,W            ; get the status register
        clrf    status              ; select bank 00
        
        bcf     SYNC_PORT, SYNC_PIN ; set the sync pin low (this is done asap)
        
        movwf   save_status         ; save the status byte
        movf    fsr, w
        movwf   save_fsr            ; and the damn fsr

;
; now determine what processing state are we in     
        movf    state, W            ; get the current state
        btfsc   status, Z_FLAG      ; is it zero ?
        goto    STATE0              ; yes !  goto state 0
        addlw   0xff                ; decrement it
        btfsc   status, Z_FLAG      ; are we zero yet ?
        goto    STATE1              ; yes !  goto state 1
        addlw   0xff                ; decrement it
        btfsc   status, Z_FLAG      ; are we zero yet ?
        goto    STATE2              ; yes !  goto state 2
        addlw   0xff                ; decrement it
        btfsc   status, Z_FLAG      ; are we zero yet ?
        goto    STATE3              ; yes !  goto state 3
        addlw   0xff                ; decrement it
        btfsc   status, Z_FLAG      ; are we zero yet ?
        goto    STATE4              ; yes !  goto state 4
// must be in state 5   

/////////////////////////////////////////////////////////
// Process state 5 :  Blank bottom of the screen
// processor cycles to here is 17
STATE5: nop
        bsf     SYNC_PORT, SYNC_PIN ; terminate the sync pulse
                                    ; nothing to do, exit straight away
        decfsz  line, F             ; decr the line counter
        goto    EXIT_INT            ; exit if we are still in state 5
        movlw   0xff
        movwf   state               ; set state=-1, other code will incr it to 0
        movlw   STATE0_CNT          ; get the line count for state 0
        goto    EXIT_STATE_CHANGE

/////////////////////////////////////////////////////////
// Process state 0 :  Vertical Sync
// nothing to do here, we even leave the sync pin low
STATE0: 
        decfsz  line, F             ; decr the line counter
        goto    EXIT_INT            ; exit if we are still in state 0
        ;
        ; this only happens once in every vert sync so we can use it to increment
        ; our clocks (i.e.  every 20mS)
        ;
        incf    skillsw_timer, f
        incf    runsw_timer, f
        incf    auto_off_timer_1, f
        btfsc   status, Z_FLAG
        incfsz  auto_off_timer_2, f
        goto S00
        ;
        ; the power down timer has timed out, put the processor into sleep
        ;
        bsf     status, 5           ; bank 01
        movlw   0xff
        movwf   5                   ; all pins set as inputs
        movwf   6                   ; all pins set as inputs
        movwf   7                   ; all pins set as inputs
        clrf    0x11                ; all analog inputs off
        clrf    0x15                ; turn off weak pullups
        bsf     option_reg, 7       ; turn off weak pullups
        bcf     status, 5           ; bank 00
        clrf    0x1f                ; turn off ADC
        clrf    0x0b                ; turn off all interrupts
        sleep                       ; this is it!  goodbye!
        
S00:    movlw   STATE1_CNT          ; setup for state 1
        goto    EXIT_STATE_CHANGE

/////////////////////////////////////////////////////////
// Process state 1 :  Blank top of the screen
// processor cycles to here is 9
STATE1: nop
        nop
        nop
        nop
        nop
        nop
        nop
        nop
        nop 
        bsf     SYNC_PORT, SYNC_PIN ; terminate the sync pulse
        // .... process state 1
        
        decfsz  line, F             ; decr the line counter
        goto    EXIT_INT            ; exit if we are still in state 1
        movlw   STATE2_CNT          ; otherwise setup for state 2
        goto    EXIT_STATE_CHANGE
    
/////////////////////////////////////////////////////////
// Process state 2 :  Single line for the top of the playing field
// processor cycles to here is 12
STATE2: nop
        nop
        nop
        nop
        nop
        nop
        BSF     SYNC_PORT, SYNC_PIN ; terminate the sync pulse
        // .... process state 2
        
        movlw   11                  ; delay for 48 instructions
M1:     addlw   0xff
        btfss   status, Z_FLAG
        goto    M1
        nop
        nop
        nop
;
; draw the top boundary line
;
        bsf     VIDEO_PORT, VIDEO_PIN   ; start the horizontal line
        movlw   39                  ; then delay for 160 instructions
M2:     addlw   0xff
        btfss   status, Z_FLAG
        goto    M2
        nop
        nop
        nop
        bcf     VIDEO_PORT, VIDEO_PIN   ; and terminate the line
        
        decfsz  line, F             ; decr the line counter
        goto    EXIT_INT            ; exit if we are still in state 2
        movlw   STATE3_CNT          ; otherwise setup for state 3
        goto    EXIT_STATE_CHANGE
    
/////////////////////////////////////////////////////////
// Process state 3 :  Main playing field
// processor cycles to here is 15
STATE3: nop
        nop
        nop 
        BSF     SYNC_PORT, SYNC_PIN ; terminate the sync pulse
        // .... process state 3
        // Left hand field margin = 12, about 50 instruction cycles from here
        // Right hand field margin = 51, about 206 instruction cycles from here
        // This gives us a field memory of 160 instruction cycles
        //      Each "pixel" is 2 instruction cycles, therefore we have 80 pixels
        //      Each byte holds 2 pixels, so we need 40 bytes
        
//      nop
        nop
        nop
        nop
;
; Generate the left bat
;
        movf    Lbat_Y, w           ; get the bat position
        subwf   line, w
        andwf   Lbat_mask, w        ; this determines the bat height
        btfsc   status, Z_FLAG
        movlw   0xf0                ; load a blank if we are NOT ready to draw
        btfss   status, Z_FLAG
        movlw   0x00                ; load a solid line if we are ready to draw
        movwf   lbuf + LBAT_X       ; and save
;
; Generate the right bat
;
        movf    Rbat_Y, w           ; get the bat position
        subwf   line, w
        andwf   Rbat_mask, w        ; this determines the bat height
        btfsc   status, Z_FLAG
        movlw   0xf0                ; load a blank if we are NOT ready to draw
        btfss   status, Z_FLAG
        movlw   0x00                ; load a solid line if we are ready to draw
        movwf   lbuf + RBAT_X       ; and save
;
; Generate the right/left margins with a gap for the goals
;
        movf    goal_top, w         ; 64 line goal centered in the vertical field
        subwf   line, w
        andwf   goal_mask, w        ; this determines the goal height
        btfss   status, Z_FLAG
        movlw   0x0f                ; load a solid line if we are ready to draw
        movwf   lbuf + LBUF_SIZE - 1  ; right margin
        btfss   practice, 1         ; test if we are in practice mode
        btfss   status, Z_FLAG
        movlw   0xf0                ; load a solid line if we are ready to draw
        movwf   lbuf                ; left margin
;
; Generate the ball
;
W3:     movf    ball_Y, w
        subwf   line, w
        andlw   0b11111100
        btfss   status, Z_FLAG
        goto    W1
        
        clrf    status
        addlw   0
        rrf     ball_X, w
        addlw   lbuf
        movwf   fsr
        rrf     ball_X, w
        btfsc   status, C_FLAG
        bsf     indf, VIDEO_PIN     ; odd value, set the lower nibble
        btfss   status, C_FLAG
        bsf     indf, VIDEO_PIN + 4 ; even value, set the high nibble
        goto    LDONE

W1:     nop
        nop     
        nop     
        nop     
        nop     
        nop     
        nop     
        nop     
        nop     
        nop     
        nop     
LDONE:      

        #include "SwapInWriteOut.c"

        clrf    indf                ; clear the ball for the next scan

        decfsz  line, f             ; decr the line counter
        goto    EXIT_INT            ; exit if we are still in state 3
        movlw   STATE4_CNT          ; otherwise setup for state 4
        goto    EXIT_STATE_CHANGE
    
/////////////////////////////////////////////////////////
// Process state 4 :  Single line for the bottom of the playing field
// processor cycles to here is 18
STATE4: //nop   
        bsf     SYNC_PORT, SYNC_PIN ; terminate the sync pulse
        
        // .... process state 4
        movlw   11                  ; delay for 48 instructions
V1:     addlw   0xff
        btfss   status, Z_FLAG
        goto    V1
        nop
        nop
        nop
;
; draw the bottom boundary line
;
        bsf     VIDEO_PORT, VIDEO_PIN   ; start the horizontal line
        movlw   39                  ; then delay for 160 instructions
V2:     addlw   0xff
        btfss   status, Z_FLAG
        goto    V2
        nop
        nop
        nop
        bcf     VIDEO_PORT, VIDEO_PIN   ; and terminate the line
        

        decfsz  line, f             ; decr the line counter
        goto    EXIT_INT            ; exit if we are still in state 4
        movlw   STATE5_CNT          ; otherwise setup for state 5
        // fall through to the exit code
        
//////////////////////////////////////////////////////      
// the last stage of going to a new state
// this is common code as all state processing code jumps here to move to the next state
EXIT_STATE_CHANGE:
        movwf   line                ; save in the counter register
        incf    state, f            ; and increment the state counter

//////////////////////////////////////////////////////  
// end of processing a horizontal line
// common code used by all state processing code
EXIT_INT:
;
; first sound the beep
;
        decfsz  beep_time, w        ; test the beep timer 
        decfsz  beep_tmp, f         ; decrement the freq count 
        goto    NOBEEP              ; skip if still in a cycle or if the timer expired 
        movlw   AUDIO_BIT           ; This is the bit of the port used for audio 
        xorwf   AUDIO_PORT          ; invert the port value 
        movf    beep_freq, w        ; reset the freq count, which is held in beep_tmp 
        movwf   beep_tmp 
        decf    beep_time, f        ; and decrement duration count 
NOBEEP: 
;
; restore fsr, status and W, return from interrupt
;
        MOVF    save_fsr, w
        MOVWF   fsr                 ; restore fsr
        SWAPF   save_status,W
        MOVWF   status              ; restore status
        SWAPF   save_w,F
        SWAPF   save_w,W            ; restore w
        BCF     intcon,2            ; re enable the interrupt.  VERY IMPORTANT
  #endasm
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////
// finally - the main processing routine
void main() {
    int i;

    // booleans used to determine collisions with the walls and bats
    short InLBatHoriz, InRBatHoriz, InRBatVert, InLBatVert, InLMargin, InRMargin, InGoal;
    short GamePause = TRUE;
    
    int skill;                                  // skill counter controlls the difficulty (bat & goal heights)
    
    // this define is complicated because the CCS compiler has trouble with arithmetic on constants
    // if the ball fails to bounce off the right bat then the cause is probably the compiler not
    // calculating the following expression correctly
    #define RTBAT_X (((LBUF_SIZE * 2) - RBAT_OFFSET * 2) * BALL_SCALE)  // horiz position of the right bat
    #define LTBAT_X LBAT_X * 2 * BALL_SCALE                             // horiz position of the left bat
    
    // set pin directions
    set_tris_A(TRIS_A); 
    #IFDEF 16F88
        set_tris_B(TRIS_B); 
    #ELIFDEF 16F688
        set_tris_C(TRIS_C); 
        wpua = ~TRIS_A;                             // turn on the weak pullups on port A
    #ENDIF

    bit_clear(option_reg, RBPU);                    // turn on the global bit controlling weak pullups
    output_high(BAT5V);                             // turn on the 5 volts for the potentiometers
    
    SetupSkill(skill = read_eeprom(0));             // initialise with the last skill level
    auto_off_timer_2 = 0;                           // reset the timeout clock
    
    setup_timer_0(RTCC_INTERNAL | RTCC_DIV_1);
    enable_interrupts(INT_TIMER0);
    enable_interrupts(GLOBAL);
    
    for (i = lbuf; i < lbuf + 40; i++) *i = 0;      // clear the line output buffer
    SetupBall();
    beep_time = 1;                                  // equivalent to turning the sound off
    
    setup_adc(ADC_CLOCK_INTERNAL );                 // Sets up the a/d mode
    setup_adc_ports(ANALOG_PINS);                   // Sets the available adc pins to be analog
    
    while(TRUE) {
        
        // check the buttons first
        if (input(SKILLSW))                         // the skill switch controlls difficulty
            skillsw_timer = 0;
        else if (skillsw_timer > 4 && skillsw_timer < 100) {        // equal to 160mS debounce time
            // the skill button has been pressed
            skillsw_timer = 216;                            // this will cause a repeat approx every second
            SetupSkill((++skill) & 0b00000111);     // four levels of skill
        }

        if (input(RUNSW))                           // the run switch starts/pauses the game
            runsw_timer = 0;
        else if (runsw_timer > 4 && runsw_timer < 9) {      // equal to 160mS debounce time
            // the skill button has been pressed
            runsw_timer = 10;                           // this will cause a repeat approx every 5 seconds
            if (tball_Y == LIMIT_Y + (BALL_SCALE + 10)) SetupBall();
            GamePause = !GamePause;
        }
        
        if (!practice) Ltbat_Y = GetBatPosition(LBAT_CHAN);     
        Rtbat_Y = GetBatPosition(RBAT_CHAN);
        
        // we do the vertical movement first
        if (!GamePause) {
            tball_Y += speed_Y;
            auto_off_timer_2 = 0;                   // and use this opportunity to reset the timeout clock
            }
            
        // check for a collision with the top and bottom
        if (!GamePause && (tball_Y >= LIMIT_Y || tball_Y <= BALL_SCALE)) {
                tball_Y += (speed_Y *= -1);
                beep_freq = BEEP_WALL_FREQ; beep_time = BEEP_WALL_TIME;
            }
            
        // check for a collision with the top or bottom of the left and right bats
        InLBatVert = ((tball_Y >= Ltbat_Y) && (tball_Y <= (Ltbat_Y + bat_size_Y)));
        InRBatVert = ((tball_Y >= Rtbat_Y) && (tball_Y <= (Rtbat_Y + bat_size_Y)));

        // and calculate if we are in the in the horizontal area of the bats
        InLBatHoriz = ((tball_X <= LTBAT_X + BALL_SCALE) && (tball_X  >=  LTBAT_X));
        InRBatHoriz = ((tball_X <= RTBAT_X + BALL_SCALE) && (tball_X  >=  RTBAT_X));
        
        // if necessary bounce the ball vertically from the bat
        if ((InLBatVert && InLBatHoriz) || (InRBatVert && InRBatHoriz)) {
            tball_Y += (speed_Y *= -1); // we have hit the bat
            beep_freq = BEEP_BAT_FREQ; beep_time = BEEP_BAT_TIME;
        }
        
        // now, the horizontal movement
        if (!GamePause) tball_X += speed_X;
        
        // we have to recalculate if we are in the in the horizontal area of the bats
        InLBatHoriz = ((tball_X <= LTBAT_X + BALL_SCALE) && (tball_X  >=  LTBAT_X));
        InRBatHoriz = ((tball_X <= RTBAT_X + BALL_SCALE) && (tball_X  >=  RTBAT_X));
        
        // if necessary bounce the ball horizontally from the bat
        if ((InLBatVert && InLBatHoriz) || (InRBatVert && InRBatHoriz)) {
            tball_X += (speed_X *= -1); // we have hit the bat
            beep_freq = BEEP_BAT_FREQ; beep_time = BEEP_BAT_TIME;
        }
        
        // now, have we hit the left or right wall ?
        InLMargin = (tball_X <= BALL_SCALE);
        InRMargin = (tball_X >= LIMIT_X - BALL_SCALE);
        
        // finally, are we in the goal ?
        InGoal = ((tball_Y > goal_top_Y) && (tball_Y <= goal_top_Y + goal_size_Y));

        // check if we are out of bounds
        if (InLMargin  || InRMargin) 
            if ((InLMargin && InGoal && !practice) || (InRMargin && InGoal)) {
                // routine to reset the game
                ball_X = 255;       // make the ball vanish
                beep_freq = BEEP_GOAL_FREQ; beep_time = BEEP_GOAL_TIME;
                tball_X = 50;
                tball_Y = LIMIT_Y + (BALL_SCALE + 10);      // place the ball out of bounds so that it "vanishes"
                GamePause = TRUE ;      // and wait for a new game
            } else {
                // we have hit the wall
                tball_X += (speed_X *= -1); 
                beep_freq = BEEP_WALL_FREQ; beep_time = BEEP_WALL_TIME;
            }

        if (state != 3) {
            ball_X = tball_X/BALL_SCALE;
            ball_Y = tball_Y/BALL_SCALE;
            Lbat_Y = Ltbat_Y/BALL_SCALE;
            Rbat_Y = Rtbat_Y/BALL_SCALE;
        }       
    }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////
// functions

//////////////////////////////////////////////////////////////////////////////////////////////////////////
// Get the position of a bat by reading the ADC
//
signed int16 GetBatPosition(int channel) {
    signed int16 t;
    
    set_adc_channel(channel);                           // Select the bat
    delay_us(10);                                       // wait for the input to settle
    t = read_adc();                                     // read the voltage
    #IFDEF 16F688
    set_tris_C(TRIS_C);                                 // the bloody ADC seems to interfere with TRIS
    #ENDIF
    t = (t - 20) * BALL_SCALE ;                         // scale to suit the field
    if (t < 0) t = 0;                                   // and keep on the field
    if (t > LIMIT_Y - bat_size_Y) t = LIMIT_Y - bat_size_Y;
    return t;
}
    
//////////////////////////////////////////////////////////////////////////////////////////////////////////
// Setup the various difficulty levels in the game
//
void SetupSkill(int skill_level) {
    switch (skill_level & 0b00000011) {
        case 0: // very easy game (for young kids)
                // the following are used in assembly language (8 bit)
                Lbat_mask = BAT_LARGE;
                Rbat_mask = BAT_LARGE;
                goal_mask = GOAL_STD;
                goal_top = (STATE3_CNT - (~GOAL_STD)) / 2;
                // the following are used in the C language (16 bit signed)
                bat_size_Y = (~BAT_LARGE) * BALL_SCALE;
                goal_size_Y = (~GOAL_STD) * BALL_SCALE;
                goal_top_Y = (LIMIT_Y - ((~GOAL_STD) * BALL_SCALE)) / 2;
                break ;
        case 1: // standard game
                // the following are used in assembly language (8 bit)
                Lbat_mask = BAT_STD;
                Rbat_mask = BAT_STD;
                goal_mask = GOAL_STD;
                goal_top = (STATE3_CNT - (~GOAL_STD)) / 2;
                // the following are used in the C language (16 bit signed)
                bat_size_Y = (~BAT_STD) * BALL_SCALE;
                goal_size_Y = (~GOAL_STD) * BALL_SCALE;
                goal_top_Y = (LIMIT_Y - ((~GOAL_STD) * BALL_SCALE)) / 2;
                break ;
        case 2: // hard game
                // the following are used in assembly language (8 bit)
                Lbat_mask = BAT_SMALL;
                Rbat_mask = BAT_SMALL;
                goal_mask = GOAL_STD;
                goal_top = (STATE3_CNT - (~GOAL_STD)) / 2;
                // the following are used in the C language (16 bit signed)
                bat_size_Y = (~BAT_SMALL) * BALL_SCALE;
                goal_size_Y = (~GOAL_STD) * BALL_SCALE;
                goal_top_Y = (LIMIT_Y - ((~GOAL_STD) * BALL_SCALE)) / 2;
                break ;
        case 3: // very hard game
                // the following are used in assembly language (8 bit)
                Lbat_mask = BAT_SMALL;
                Rbat_mask = BAT_SMALL;
                goal_mask = GOAL_LARGE;
                goal_top = (STATE3_CNT - (~GOAL_LARGE)) / 2;
                // the following are used in the C language (16 bit signed)
                bat_size_Y = (~BAT_SMALL) * BALL_SCALE;
                goal_size_Y = (~GOAL_LARGE) * BALL_SCALE;
                goal_top_Y = (LIMIT_Y - ((~GOAL_LARGE) * BALL_SCALE)) / 2;
                break ;
    }
    
    if (skill_level & 0b00000100) {
        // practice play
        practice = 0xff;
        Lbat_mask = 0b11111111;                 // reduces the bat to one pixel
        Ltbat_Y = LIMIT_Y + (BALL_SCALE + 10);  // and places it outside of the playing field
    } else {
        // standard play
        practice = 0;
    }
    
    // write the new value to non eeprom memory
    if (read_eeprom(0) != (skill_level & 0b00000111)) write_eeprom(0, (skill_level & 0b00000111));
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////
// Setup the ball for the beginning of a game
//
void SetupBall() {
    int i;
    
    tball_X = LIMIT_X/2; tball_Y = LIMIT_Y/2;       // place the ball in the centre
    speed_X = speed_Y = BALL_START_SPEED;           // and set a standard speed
    i = auto_off_timer_1 ;                                  // use one of our clocks as a random value
    if (i & 0b00000001) speed_X *= -1;              // and set random directions
    if (i & 0b00000010) speed_Y *= -1;
}
